--[[
Name: TankPoints Calculator
Revision: $Revision: 2 $
Developed by: Whitetooth@Cenarius (hotdogee@ګiSbaha)
Description: Interactive Calculator for TankPoints
]]
-- Librarys
local Compost = AceLibrary("Compost-2.0")
local L = AceLibrary("AceLocale-2.2"):new("TankPoints")

-- Calc Table
TPCalc = {}
TPCalc.playerDT = {}

-----------
-- Tools --
-----------
-- copyTable from AceDB
local function copyTable(to, from)
	setmetatable(to, nil)
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable({}, k)
		end
		if type(v) == "table" then
			v = copyTable({}, v)
		end
		to[k] = v
	end
	setmetatable(to, from)
	return to
end

-------------------------------
-- TankPointsCalculatorFrame --
-------------------------------
function TankPointsCalculatorFrame_OnLoad()
	-- Esc closes the window
	tinsert(UISpecialFrames, this:GetName())
	
	-- Set title text
	TankPointsCalculatorFrame_HeaderText:SetText(L["TankPoints Calculator"])
	-- Set drag frame tooltip
	TankPointsCalculatorFrame_DragFrame.tooltip = L["Left click to drag\nRight click to reset position"]
	-- Set button text
	TankPointsCalculatorFrame_ResetButton:SetText(L["Reset"])
	TankPointsCalculatorFrame_CloseButton:SetText(L["Close"])
	-- Set option frame box title
	TPCResultsFrameTitle:SetText(L["Results"])
	TPCCombatTableFrameTitle:SetText(L["Combat Table"])
	TPCPlayerStatsFrameTitle:SetText(L["Player Stats"])
	TPCMobStatsFrameTitle:SetText(L["Mob Stats"])
	-- Set label text
	TPCalc:SetLabels()
	
	-- Set tooltips
	TPCMobStats2.tooltip = L["Mob Damage before DR"]
	
	-- Register events
	this:RegisterEvent("UNIT_LEVEL")
	this:RegisterEvent("UNIT_RESISTANCES")
	this:RegisterEvent("UNIT_STATS")
	this:RegisterEvent("UNIT_DEFENSE")
	this:RegisterEvent("UNIT_MAXHEALTH")
	this:RegisterEvent("UNIT_AURA")
	this:RegisterEvent("UNIT_INVENTORY_CHANGED")
end

function TankPointsCalculatorFrame_OnEvent(event)
	-- Do nothing if Calculator frame is not visable
	if (not this:IsVisible()) then
		return
	end
	-- Do nothing if event target is not player
	if (not arg1 == "player") then
		return
	end
	-- Update stuff
	TPCalc:UpdateBase()
	TPCalc:UpdateResults()
end

function TankPointsCalculatorFrame_OnShow()
	TPCalc:UpdateBase()
	TPCalc:UpdateResults()
end

-- VariableFrame
function TankPointsCalculatorVariables_IncrementButton_OnClick(button, down)
	local inputBox = getglobal(this:GetParent():GetName().."_InputEditBox")
	inputBox:SetNumber(inputBox:GetNumber() + 1)
	inputBox:ClearFocus()
end

function TankPointsCalculatorVariables_DecrementButton_OnClick(button, down)
	local inputBox = getglobal(this:GetParent():GetName().."_InputEditBox")
	inputBox:SetNumber(inputBox:GetNumber() - 1)
	inputBox:ClearFocus()
end

function TankPointsCalculatorVariables_InputEditBox_OnTextChanged()
	if (this:GetNumber() > 0) then
		this:SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (this:GetNumber() < 0) then
		this:SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		this:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	TPCalc:UpdateResults()
end

TPCalc.LabelText = {
	{-- TPCResults1
		L["TankPoints"],
		L["Total Reduction"]..L["(%)"],
	},
	{-- TPCCombatTable1
		MISS..L["(%)"],
		DODGE..L["(%)"],
		PARRY..L["(%)"],
		BLOCK..L["(%)"],
		L["Crit"]..L["(%)"],
		L["Crushing"]..L["(%)"],
		L["Hit"]..L["(%)"],
	},
	{-- TPCPlayerStats1
		SPELL_STAT1_NAME, -- "Strength"
		SPELL_STAT2_NAME, -- "Agility"
		SPELL_STAT3_NAME, -- "Stamina"
		L["Max Health"],
		"["..ARMOR.." - "..L["Items"].."]",
		"["..ARMOR.."]",
		"["..COMBAT_RATING_NAME2.."]", -- "Defense Rating"
		"["..DEFENSE.."]",
		"["..COMBAT_RATING_NAME3.."]", -- "Dodge Rating"
		"["..DODGE..L["(%)"].."]",
		"["..COMBAT_RATING_NAME4.."]", -- "Parry Rating"
		"["..PARRY..L["(%)"].."]",
		"["..COMBAT_RATING_NAME5.."]", -- "Block Rating"
		"["..BLOCK..L["(%)"].."]",
		L["Block Value"],
		COMBAT_RATING_NAME15, -- "Resilience"
	},
	{-- TPCMobStats1
		L["Mob Level"],
		L["Mob Damage"],
	},
}
-- Set label text
function TPCalc:SetLabels()
	for i, text in ipairs(self.LabelText[1]) do
		getglobal("TPCResults"..i.."_LabelText"):SetText(text)
	end
	for i, text in ipairs(self.LabelText[2]) do
		getglobal("TPCCombatTable"..i.."_LabelText"):SetText(text)
	end
	for i, text in ipairs(self.LabelText[3]) do
		getglobal("TPCPlayerStats"..i.."_LabelText"):SetText(text)
	end
	for i, text in ipairs(self.LabelText[4]) do
		getglobal("TPCMobStats"..i.."_LabelText"):SetText(text)
	end
end

function TankPointsCalculatorFrame_ResetButton_OnClick(button, down)
	for i, _ in ipairs(TPCalc.LabelText[3]) do
		local inputBox = getglobal("TPCPlayerStats"..i.."_InputEditBox")
		inputBox:SetText("0")
		inputBox:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	for i, _ in ipairs(TPCalc.LabelText[4]) do
		local inputBox = getglobal("TPCMobStats"..i.."_InputEditBox")
		inputBox:SetText("0")
		inputBox:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
end

-- Update base data
function TPCalc:UpdateBase()
	TankPoints:GetPlayerDataTable(self.playerDT)
	TankPoints:GetTankPoints(self.playerDT)
end

function TPCalc:UpdateResults()
	local newDT = Compost:Acquire()
	TankPoints:GetPlayerDataTable(newDT, TP_MELEE)
	local prefix
	--------------------
	-- Get input data --
	--------------------
	-- input data is writen in the changes table
	local changes = Compost:Acquire()
	local inputEditBox = "_InputEditBox"
	local currentText = "_CurrentText"
	local resultText = "_ResultText"
	local differenceText = "_DifferenceText"
	local originalStatText = "_OriginalStatText"
	local newStatText = "_NewStatText"
	
	local current, new, diff
	
	prefix = "TPCPlayerStats"
	local i = 1
	-- Strength
	changes.str = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Agility
	changes.agi = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Stamina
	changes.sta = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Max Health
	changes.playerHealth = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Armor (Items)
	changes.armorFromItems = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Armor
	changes.armor = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Defense Rating
	changes.defense = 0
	diff = getglobal(prefix..i..inputEditBox):GetNumber()
	if diff then
		-- Ratings are floored in game
		current = GetCombatRating(CR_DEFENSE_SKILL)
		changes.defense = floor(ReverseRating(current + diff, RB_DEFENSE_SKILL, newDT.playerLevel)) - floor(ReverseRating(current, RB_DEFENSE_SKILL, newDT.playerLevel))
	end
	i = i + 1
	-- Defense
	changes.defense = changes.defense + getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Dodge Rating
	changes.dodgeChance = 0
	diff = getglobal(prefix..i..inputEditBox):GetNumber()
	if diff then
		changes.dodgeChance = ReverseRating(diff, RB_DODGE, newDT.playerLevel) * 0.01
	end
	i = i + 1
	-- Dodge
	changes.dodgeChance = changes.dodgeChance + getglobal(prefix..i..inputEditBox):GetNumber() * 0.01
	i = i + 1
	-- Parry Rating
	changes.parryChance = 0
	diff = getglobal(prefix..i..inputEditBox):GetNumber()
	if diff then
		changes.parryChance = ReverseRating(diff, RB_PARRY, newDT.playerLevel) * 0.01
	end
	i = i + 1
	-- Parry
	changes.parryChance = changes.parryChance + getglobal(prefix..i..inputEditBox):GetNumber() * 0.01
	i = i + 1
	-- Block Rating
	changes.blockChance = 0
	diff = getglobal(prefix..i..inputEditBox):GetNumber()
	if diff then
		changes.blockChance = ReverseRating(diff, RB_BLOCK, newDT.playerLevel) * 0.01
	end
	i = i + 1
	-- Block
	changes.blockChance = changes.blockChance + getglobal(prefix..i..inputEditBox):GetNumber() * 0.01
	i = i + 1
	-- Block Value
	changes.blockValue = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- Resilience
	changes.resilience = getglobal(prefix..i..inputEditBox):GetNumber()
	
	prefix = "TPCMobStats"
	i = 1
	-- mobLevel
	changes.mobLevel = getglobal(prefix..i..inputEditBox):GetNumber()
	i = i + 1
	-- mobDamage
	changes.mobDamage = getglobal(prefix..i..inputEditBox):GetNumber()
	
	----------------
	-- AlterTable --
	----------------
	TankPoints:AlterDataTable(newDT, changes)
	
	------------------------------
	-- Calculate new TankPoints --
	------------------------------
	TankPoints:GetTankPoints(newDT, TP_MELEE)
	
	------------------
	-- Display data --
	------------------
	--TPCPlayerStats1_OriginalStatText
	--TPCPlayerStats1_NewStatText
	--TPCResults1_CurrentText
	--TPCResults1_DifferenceText
	--TPCResults1_ResultText
	
	
	-------------------
	-- Results Frame --
	-------------------
	prefix = "TPCResults"
	-- TankPoints
	i = 1
	current = floor(self.playerDT.tankPoints[TP_MELEE])
	new = floor(newDT.tankPoints[TP_MELEE])
	getglobal(prefix..i..currentText):SetText(current)
	getglobal(prefix..i..resultText):SetText(new)
	getglobal(prefix..i..differenceText):SetText(format("%d", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Damage Reduction
	i = i + 1
	current = floor(self.playerDT.totalReduction[TP_MELEE] * 100 * 100) / 100
	new = floor(newDT.totalReduction[TP_MELEE] * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	------------------------
	-- Combat Table Frame --
	------------------------
	prefix = "TPCCombatTable"
	-- mobMissChance
	i = 1
	current = floor(self.playerDT.mobMissChance * 100 * 100) / 100
	new = floor(newDT.mobMissChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- dodgeChance
	i = i + 1
	current = floor(self.playerDT.dodgeChance * 100 * 100) / 100
	new = floor(newDT.dodgeChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- parryChance
	i = i + 1
	current = floor(self.playerDT.parryChance * 100 * 100) / 100
	new = floor(newDT.parryChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- blockChance
	i = i + 1
	current = floor(self.playerDT.blockChance * 100 * 100) / 100
	new = floor(newDT.blockChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- mobCritChance
	i = i + 1
	current = floor(self.playerDT.mobCritChance * 100 * 100) / 100
	new = floor(newDT.mobCritChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- mobCrushChance
	i = i + 1
	current = floor(self.playerDT.mobCrushChance * 100 * 100) / 100
	new = floor(newDT.mobCrushChance * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- mobHitChance
	i = i + 1
	current = floor(max(0, (1 - self.playerDT.mobCrushChance - self.playerDT.mobCritChance - self.playerDT.blockChance - self.playerDT.parryChance - self.playerDT.dodgeChance - self.playerDT.mobMissChance)) * 100 * 100) / 100
	new = floor(max(0, (1 - newDT.mobCrushChance - newDT.mobCritChance - newDT.blockChance - newDT.parryChance - newDT.dodgeChance - newDT.mobMissChance)) * 100 * 100) / 100
	getglobal(prefix..i..currentText):SetText(format("%.2f", current))
	getglobal(prefix..i..resultText):SetText(format("%.2f", new))
	getglobal(prefix..i..differenceText):SetText(format("%.2f", (new - current)))
	if (new < current) then
		getglobal(prefix..i..resultText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new > current) then
		getglobal(prefix..i..resultText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..resultText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		getglobal(prefix..i..differenceText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	------------------------
	-- Player Stats Frame --
	------------------------
	prefix = "TPCPlayerStats"
	-- Strength
	i = 1
	_, current = UnitStat("player", 1)
	new = floor(current + changes.str)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Agility
	i = i + 1
	_, current = UnitStat("player", 2)
	new = floor(current + changes.agi)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Stamina
	i = i + 1
	_, current = UnitStat("player", 3)
	new = floor(current + changes.sta)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Max Health
	i = i + 1
	current = self.playerDT.playerHealth
	new = floor(newDT.playerHealth)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Armor (Items)
	i = i + 1
	current = self.playerDT.armor
	new = floor(newDT.armor)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Armor
	i = i + 1
	current = self.playerDT.armor
	new = floor(newDT.armor)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Defense Rating
	i = i + 1
	current = GetCombatRating(CR_DEFENSE_SKILL)
	new = floor(current + getglobal(prefix..i..inputEditBox):GetNumber())
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Defense
	i = i + 1
	current = self.playerDT.defense
	new = floor(newDT.defense)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Dodge Rating
	i = i + 1
	current = GetCombatRating(CR_DODGE)
	new = floor(current + getglobal(prefix..i..inputEditBox):GetNumber())
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Dodge
	i = i + 1
	current = floor(self.playerDT.dodgeChance * 100 * 100) / 100
	new = floor(newDT.dodgeChance * 100 * 100) / 100
	getglobal(prefix..i..originalStatText):SetText(format("%.2f", current))
	getglobal(prefix..i..newStatText):SetText(format("%.2f", new))
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Parry Rating
	i = i + 1
	current = GetCombatRating(CR_PARRY)
	new = floor(current + getglobal(prefix..i..inputEditBox):GetNumber())
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Parry
	i = i + 1
	current = floor(self.playerDT.parryChance * 100 * 100) / 100
	new = floor(newDT.parryChance * 100 * 100) / 100
	getglobal(prefix..i..originalStatText):SetText(format("%.2f", current))
	getglobal(prefix..i..newStatText):SetText(format("%.2f", new))
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Block Rating
	i = i + 1
	current = GetCombatRating(CR_BLOCK)
	new = floor(current + getglobal(prefix..i..inputEditBox):GetNumber())
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Block
	i = i + 1
	current = floor(self.playerDT.blockChance * 100 * 100) / 100
	new = floor(newDT.blockChance * 100 * 100) / 100
	getglobal(prefix..i..originalStatText):SetText(format("%.2f", current))
	getglobal(prefix..i..newStatText):SetText(format("%.2f", new))
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Block Value
	i = i + 1
	current = self.playerDT.blockValue
	new = floor(newDT.blockValue)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- Resilience
	i = i + 1
	current = self.playerDT.resilience
	new = floor(newDT.resilience)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	---------------------
	-- Mob Stats Frame --
	---------------------
	prefix = "TPCMobStats"
	-- mobLevel
	i = 1
	current = self.playerDT.mobLevel
	new = floor(newDT.mobLevel)
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	-- mobDamage
	i = i + 1
	current = floor(TankPoints:GetMobDamage(self.playerDT.mobLevel))
	new = floor(TankPoints:GetMobDamage(newDT.mobLevel) + getglobal(prefix..i..inputEditBox):GetNumber())
	getglobal(prefix..i..originalStatText):SetText(current)
	getglobal(prefix..i..newStatText):SetText(new)
	if (new > current) then
		getglobal(prefix..i..newStatText):SetTextColor(GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
	elseif (new < current) then
		getglobal(prefix..i..newStatText):SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g, RED_FONT_COLOR.b)
	else
		getglobal(prefix..i..newStatText):SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	end
	
	--------------
	-- Clean up --
	--------------
	Compost:Reclaim(newDT, 2)
	Compost:Reclaim(changes)
end